import logging
from ..static import LOGGER_PRINT_LEVEL
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(LOGGER_PRINT_LEVEL)

from pprint import pprint

import torch
from torch.utils.data import DataLoader

# Mask variable
def _mask(prev_generated_seq, config):
    prev_mask = torch.eq(prev_generated_seq, 1).type(torch.float32)
    lengths = torch.argmax(prev_mask,dim=1)
    max_len = prev_generated_seq.size(1)
    mask = []
    for i in range(prev_generated_seq.size(0)):
        if lengths[i] == 0:
            mask_line = [0] * max_len
        else:
            mask_line = [0] * lengths[i].item()
            mask_line.extend([1] * (max_len - lengths[i].item()))
        mask.append(mask_line)
    mask = torch.ByteTensor(mask)
    if config.cuda_is_available:
        mask = mask.cuda()
    return prev_generated_seq.data.masked_fill_(mask, 0.)

def train_batch(model, criterion, optimizer, vocab_size,
                input_variable, input_lengths, target_variable, 
                teacher_forcing_ratio, config):
    loss_list = []
    # Forward propagation
    prev_generated_seq = None
    target_variable_reshaped = target_variable[:, 1:].contiguous().view(-1)

    for i in range(config.num_of_exams):
        decoder_outputs, _, other = \
            model(input_variable, prev_generated_seq, input_lengths,
                   target_variable, teacher_forcing_ratio)

        decoder_outputs_reshaped = decoder_outputs.view(-1, vocab_size)
        lossi = criterion(decoder_outputs_reshaped, target_variable_reshaped)
        loss_list.append(lossi.item())
        model.zero_grad()
        lossi.backward(retain_graph=True)
        torch.nn.utils.clip_grad_norm_(model.parameters(), config.max_grad_norm)
        optimizer.step()
        prev_generated_seq = torch.squeeze(torch.topk(decoder_outputs, 1, dim=2)[1]).view(-1, decoder_outputs.size(1))
        prev_generated_seq = _mask(prev_generated_seq, config)
    return loss_list


def train_epoches(dataset, model, criterion, optimizer, 
                    vocab_size, teacher_forcing_ratio, 
                    config):
    train_loader = DataLoader(dataset, config.batch_size)
    model.train(True)
    prev_epoch_loss_list = [100] * config.num_of_exams
    for epoch in range(1, config.epochs + 1):
        epoch_examples_total = 0
        epoch_loss_list = [0] * config.num_of_exams
        for batch_idx, (source, target, input_lengths) in enumerate(train_loader):
            input_variables = source
            target_variables = target
            # train model
            loss_list = train_batch(model, criterion, optimizer, vocab_size,
                                        input_variables, input_lengths.tolist(),
                                        target_variables, teacher_forcing_ratio,
                                        config)
            # Record average loss
            num_examples = len(source)
            epoch_examples_total += num_examples
            for i in range(config.num_of_exams):
                epoch_loss_list[i] += loss_list[i] * num_examples

        for i in range(config.num_of_exams):
            epoch_loss_list[i] /= float(epoch_examples_total)

        log_msg = "Finished epoch %d with losses:" % epoch
        logger.debug(log_msg)
        pprint(epoch_loss_list)
        if prev_epoch_loss_list[:-1] < epoch_loss_list[:-1]:
            break
        else:
            prev_epoch_loss_list = epoch_loss_list[:]



import logging
from ..static import LOGGER_PRINT_LEVEL
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(LOGGER_PRINT_LEVEL)

import torch.nn as nn
import torch.nn.functional as F

class WritingEditingNetwork(nn.Module):
    def __init__(self, encoder_title, encoder, decoder,
                    decode_function=F.log_softmax) -> None:
        super().__init__()
        self.encoder_title = encoder_title
        self.encoder = encoder
        self.decoder = decoder
        self.decode_function = decode_function
    
    def flatten_parameters(self) -> None:
        self.encoder.rnn.flatten_parameters()
        self.decoder.rnn.flatten_parameters()

    def forward(self, input_variable, prev_generated_seq = None, 
                    input_lengths = None, target_variable = None,
                    teacher_forcing_ratio = 0):
        encoder_outputs, encoder_hidden = self.encoder_title(input_variable,
                                            input_lengths)
        if prev_generated_seq is None:
            pg_encoder_states = None
            pg_encoder_hidden = None
        else:
            pg_encoder_states, pg_encoder_hidden = self.encoder(
                                                        prev_generated_seq
                                                    )
        result = self.decoder(inputs = target_variable,
                                encoder_hidden = encoder_hidden,
                                encoder_outputs = encoder_outputs,
                                pg_encoder_states = pg_encoder_states,
                                function = self.decode_function,
                                teacher_forcing_ratio = teacher_forcing_ratio)
        return result



import logging
from ..static import LOGGER_PRINT_LEVEL
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(LOGGER_PRINT_LEVEL)

import os
from typing import Union, List
from collections import Counter
import numpy as np

import torch
from torch.utils.data import Dataset

class Data(object):
    def __init__(self, filepath: str = '') -> None:
        self.filepath = filepath        
        self.data = [] # [[title, abstract], [title, abstract]]
        self.train_idx = None
        self.validation_idx = None
        self.test_idx = None

    def load(self) -> bool:
        logger.debug('loading dataset...')
        if self.filepath:
            logger.debug('filepath: {filepath}'.format(filepath=self.filepath))
            with open(self.filepath, 'r') as f:
                lines = f.readlines()            
            logger.info('# of lines in raw data: {num_of_lines}'.format(num_of_lines=len(lines)))

            counter = 0     
            for i in range(0, len(lines), 3):
                if (len(lines[i].strip()) > 0) and (len(lines[i+1].strip()) > 0):
                    self.data.append((lines[i], lines[i+1]))                    
                else:
                    counter += 1
            logger.info('# of lines in loaded data: {num_of_lines}'.format(num_of_lines=len(self.data)))
            logger.info('# of lines excluded: {num_of_lines}'.format(num_of_lines=counter))
            logger.debug('loading dataset completed.')
            return True
    
    def train_test_split(self, test_size: float, train_size: float, 
                            seed: Union[int, None] = None) -> None:
        logger.debug('performing train test split...')
        total = len(self.data)
        
        self.train_idx = int(train_size * total)
        self.test_idx  = total  
        if (test_size + train_size) == 1.0:
            self.validation_idx = None                        
        else:
            test_count = int(test_size * total)
            validation_count = total - self.train_idx - test_count
            self.validation_idx = self.train_idx + validation_count
        
        logger.info('total # of rows: {num_of_rows}'.format(num_of_rows=total))
        logger.info('# of training data: {num_of_rows}'.format(num_of_rows=self.train_idx))
        logger.info('# of validation data: {num_of_rows}'.format(num_of_rows=(self.validation_idx-self.train_idx)))
        logger.info('# of test data: {num_of_rows}'.format(num_of_rows=(self.test_idx-self.validation_idx)))                    
        
        if self.data:
            if seed:
                random_state = np.random.RandomState(seed)
            else:
                random_state = np.random.RandomState()
            random_state.shuffle(self.data)
            logger.debug('performing train test split completed.')
            return
        else:
            logger.error('Please load data first before performing train test split.')
            return

    def save(self, directory: str) -> bool:
        with open(os.path.join(directory, 'data.dat'), 'w') as f:
            _ = [''.join(row) + '\n' for row in self.data]            
            f.write(
                ''.join(_)
            )
        with open(os.path.join(directory, 'train.dat'), 'w') as f:
            _ = [''.join(row) + '\n' for row in self.data[:self.train_idx]]            
            f.write(
                ''.join(_)
            ) 
        if self.validation_idx:
            with open(os.path.join(directory, 'validation.dat'), 'w') as f:
                _ = [''.join(row) + '\n' for row in self.data[self.train_idx:self.validation_idx]]                
                f.write(
                    ''.join(_)
                ) 
            with open(os.path.join(directory, 'test.dat'), 'w') as f:
                _ = [''.join(row) + '\n' for row in self.data[self.validation_idx:self.test_idx]]                
                f.write(
                    ''.join(_)
                )                 
        else:
            with open(os.path.join(directory, 'test.dat'), 'w') as f:
                _ = [''.join(row) + '\n' for row in self.data[self.train_idx:self.test_idx]]                
                f.write(
                    ''.join(_)
                )             

class Vectorizer(object):
    def __init__(self, max_words: Union[int, None] = None,
                    min_frequency: Union[int, None] = None,
                    start_end_tokens: bool = True,
                    max_len: Union[int, None] = None):
        
        self.max_words = max_words
        self.min_frequency = min_frequency
        self.start_end_tokens = start_end_tokens
        self.max_len = max_len

        self.vocabulary = dict()
        self.vocabulary_excluded = dict()
        self.vocabulary_size = 0 
        self.word2idx = dict()
        self.idx2word = dict()

    def _find_max_sentence_length(self, corpus: List, template: bool) -> None:
        logger.debug('finding max length from corpus...')
        if not template:
            self.max_len = max(len(sent) 
                                for document in corpus
                                for sent in document)
        else:
            self.max_len = max(len(sent)
                                for sent in corpus)
            
        logger.info('max length without start and end tokens: {max_len}'.format(max_len=self.max_len))        
        if self.start_end_tokens:
            self.max_len += 2
            logger.info('max length with start and end tokens: {max_len}'.format(max_len=self.max_len))
        
        logger.debug('finding max length from corpus completed.')
        
    def _build_vocabulary(self, corpus: List, template: bool) -> None:
        """
        Instantiates `Vectorizer.vocabulary` with a dictionary
        of {word:freq}.
        """
        logger.debug('building vocabulary...')
        if not template:
            vocabulary = Counter(word for document in corpus 
                                 for sent in document
                                 for word in sent)
            
        else:
            vocabulary = Counter(word for sent in corpus
                                    for word in sent)        
        
        logger.info('# of words in vocabulary: {num_of_words}'.format(num_of_words=len(vocabulary)))

        if self.max_words:
            vocabulary = {
                word:freq for word, freq in vocabulary.most_common(self.max_words)
            }
            logger.info('# of words in vocabulary (capped): {num_of_words}'.format(num_of_words=len(vocabulary)))
        
        if self.min_frequency:
            for word, freq in vocabulary.items():
                if freq >= self.min_frequency:
                    self.vocabulary[word] = freq
                else:
                    self.vocabulary_excluded[word] = freq

        logger.info('# of words meeting min freq requirement: {num_of_words}'.format(num_of_words=len(self.vocabulary)))
        logger.info('# of words excluded: {num_of_words}'.format(num_of_words=len(self.vocabulary_excluded)))

        self.vocabulary_size = len(self.vocabulary) + 2 # why?
        if self.start_end_tokens:
            self.vocabulary_size += 2
       
        logger.info('# of words in vocabulary (final): {num_of_words}'.format(num_of_words=len(self.vocabulary)))
        logger.info('vocabulary_size: {vocabulary_size}'.format(vocabulary_size=self.vocabulary_size))

        logger.debug('building vocabulary completed.')

    def _build_word_index(self) -> None:
        """
        Instantiates `Vectorizer.word2idx` and `Vectorizer.idx2word`
        with each word in `Vectorizer.vocabulary` and its associated
        index.
        """
        logger.debug('building word index...')
        self.word2idx['<UNK>'] = 3
        self.word2idx['<PAD>'] = 0

        if self.start_end_tokens:
            self.word2idx['<EOS>'] = 1
            self.word2idx['<SOS>'] = 2
        
        offset = len(self.word2idx)
        for idx, word in enumerate(self.vocabulary):
            self.word2idx[word] = idx + offset
        
        self.idx2word = {idx:word for word, idx in self.word2idx.items()}
        logger.debug('building word index completed.')

    def transform_sentence(self, sentence: str) -> List:
        vector = [self.word2idx.get(word, self.word2idx['<UNK>']) 
                    for word in sentence]
        return vector

    def fit(self, corpus: List, template: bool = False) -> None:
        logger.debug('fitting corpus...')
        if not self.max_len:
            self._find_max_sentence_length(corpus, template)            
        self._build_vocabulary(corpus, template)
        self._build_word_index()        
        logger.debug('fitting corpus completed.')

    def transform(self, corpus: List, template: bool = False) -> List:
        """
        Converts corpus to list of indices.
        """
        logger.debug('creating word index vector...')
        vcorpus = []
        if not template:
            for document in corpus:
                vcorpus.append([self.transform_sentence(sentence) for sentence in document])
        else:
            vcorpus.extend([self.transform_sentence(sentence) for sentence in corpus])
        
        logger.debug('creating word index vector completed.')
        return vcorpus

class Abstracts(Dataset):
    def __init__(self, use_cuda: bool, max_len: int = 200) -> None:
        self.max_len = max_len # allowed max length of abstracts 
        self.use_cuda = use_cuda

        self.head_len = 0 # actual max length of titles
        self.abs_len = 0 # actual max length of abstracts, capped at `max_len`
        
        self.data = None
        #self.corpus = self._read_corpus(path)
        #self.data = self._vectorize_corpus()
        #self._initalcorpus()

    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index):
        len_s, len_t, source, target = self.data[index]
        source = torch.LongTensor(source)
        target = torch.LongTensor(target)

        if self.use_cuda:
            source = source.cuda()
            target = target.cuda()
        
        return source, target, len_s

    def pad_sentence_vector(self, vector: List, max_len: int) -> List:
        padding = max_len - len(vector)
        vector.extend([0] * padding)
        return vector
    
    def _tokenize_word(self, sentence: str) -> List:  
        """
        Simple tokenization just by splitting by space.
        """      
        return [word for word in sentence.split() if word]        

    def _read_corpus(self, data: Data) -> List:
        """
        Converts data from [[title, abstract], [title, abstract], ...]
        to [[
                [title_word_1, title_word_2, ...], [abstract_word_1, abstract_word_2, ...]
            ],
            [
                [title_word_1, title_word_2, ...], [abstract_word_1, abstract_word_2, ...]
            ],
            ...
        ]
        """
        logger.debug('reading corpus...')
        logger.info('# of rows in dataset: {num_of_rows}'.format(num_of_rows=len(data.data)))
        corpus = []        
        counter = 0
        for pair in data.data:            
            title = self._tokenize_word(pair[0])
            abstract = self._tokenize_word(pair[1])                
            if title and abstract:                
                corpus.append([title, abstract])
            else:
                counter += 1

        logger.info('# of rows in corpus: {num_of_rows}'.format(num_of_rows=len(corpus)))
        logger.info('# of rows excluded: {num_of_rows}'.format(num_of_rows=counter))
        
        logger.debug('reading corpus completed.')
        return corpus
    
    def _vectorize_corpus(self, corpus: List, vectorizer: Vectorizer):
        """
        Converts corpus to list of indices.
        """
        logger.debug('vectorizing corpus...')
        if not vectorizer.vocabulary:            
            vectorizer.fit(corpus)            
        transformed = vectorizer.transform(corpus)
        logger.debug('vectorizing corpus completed.')
        return transformed
    
    def _initalcorpus(self, vectorized_corpus: List,
                        vectorizer: Vectorizer):
        """
        Apply padding to titles and abstracts with zeros to 
        make them of equal length.
        """
        # determine max length of titles and abstracts
        # pad titles and abstracts
        old = []
        for i in vectorized_corpus:
            source = i[0] # list of indices for title
            target = i[1] # list of indices for abstract
            if len(source) > self.head_len:
                self.head_len = len(source)

            if len(target) <= self.max_len:
                if len(target) > self.abs_len:
                    self.abs_len = len(target)
            else:
                target = target[:self.max_len-1]
                target.append(vectorizer.word2idx['<EOS>'])
                self.abs_len = len(target)
            old.append((source[1:-1], target))
        old.sort(key=lambda x: len(x[0]), reverse=True) # descending
        
        corpus = []
        for source, target in old:
            team = [
                        len(source), 
                        len(target), 
                        self.pad_sentence_vector(source, self.head_len), 
                        self.pad_sentence_vector(target, self.abs_len)
                    ]
            corpus.append(team)        
        self.data = corpus

    def process_corpus(self, data: Data, vectorizer: Vectorizer) -> None:
        corpus = self._read_corpus(data)
        corpus = self._vectorize_corpus(corpus, vectorizer)
        self._initalcorpus(corpus, vectorizer)




import logging
import os

LOGGER_PRINT_LEVEL = logging.DEBUG
DIRECTORY = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DIRECTORY_DATA = os.path.join(DIRECTORY, 'data')
DIRECTORY_OUTPUT = os.path.join(DIRECTORY, 'output')

class Config(object):
    seed = 1111
    cudnn_benchmark = True
    cuda_is_available = False

    # vectorizer configs
    max_words = None # determined from dataset if None
    min_freq = 5 # no filter if None 
    start_end_tokens = True
    max_len = None # determined from dataset if None

    # embedding configs
    emsize = 512
    embedding_size = 512

    # model configs
    cell = "GRU"
    learning_rate = 0.001
    num_of_exams = 3
    
    nlayers = 1    
    epochs = 10
    batch_size = 64
    dropout = 0
    bidirectional = True
    max_grad_norm = 10   

    relative_data_path = '/data/train.dat'
    relative_dev_path = '/data/dev.dat'
    relative_gen_path = '/data/fake%d.dat'
     
    

class ConfigTest(Config):    
    emsize = 3
    lr = 1
    epochs = 3
    batch_size = 2
    relative_data_path = '/data/haha.dat'
    relative_dev_path = '/data/haha.dat'
    relative_gen_path = '/data/fake%d.dat'
    max_grad_norm = 1
    min_freq = 0



import logging
from ..static import LOGGER_PRINT_LEVEL
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(LOGGER_PRINT_LEVEL)

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

class Attention(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.linear_out = nn.Linear(dim*2, dim)

    def forward(self, output, context):
        batch_size = output.size(0)
        hidden_size = output.size(2)
        input_size = context.size(1)

        # (batch, out_len, dim) * (batch, in_len, dim) -> (batch, out_len, in_len)
        attn = torch.bmm(output, context.transpose(1, 2))
        mask = torch.eq(attn, 0).data.byte()
        attn.data.masked_fill_(mask, -float('inf'))
        attn = F.softmax(attn.view(-1, input_size), dim=1).view(batch_size, -1, input_size)

        # (batch, out_len, in_len) * (batch, in_len, dim) -> (batch, out_len, dim)
        mix = torch.bmm(attn, context)

        # concat -> (batch, out_len, 2*dim)
        combined = torch.cat((mix, output), dim=2)

        # output -> (batch, out_len, dim)
        output = torch.tanh(self.linear_out(combined.view(-1, 2 * hidden_size))).view(batch_size, -1, hidden_size)

        if not output.is_contiguous():
            output = output.contiguous()

        return output, attn


import logging
from ..static import LOGGER_PRINT_LEVEL
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(LOGGER_PRINT_LEVEL)

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import BaseRNN
from .attention import Attention

class Gate(nn.Module):
    def __init__(self, hidden_size: int) -> None:
        super().__init__()
        self.hidden_size = hidden_size

        self.wrx = nn.Linear(hidden_size, hidden_size)
        self.wrh = nn.Linear(hidden_size, hidden_size)
        self.wix = nn.Linear(hidden_size, hidden_size)
        self.wih = nn.Linear(hidden_size, hidden_size)
        self.wnx = nn.Linear(hidden_size, hidden_size)
        self.wnh = nn.Linear(hidden_size, hidden_size)

    def forward(self, title, pg):
        r_gate = torch.sigmoid(self.wrx(title) + self.wrh(pg))
        i_gate = torch.sigmoid(self.wix(title) + self.wih(pg))
        n_gate = torch.tanh(self.wnx(title) + torch.mul(r_gate, self.wnh(pg)))
        result = torch.mul(i_gate, pg) + torch.mul(torch.add(-i_gate, 1), n_gate)
        return result
    
class DecoderRNNFB(BaseRNN):

    KEY_ATTN_SCORE = 'attention_score'
    KEY_LENGTH = 'length'
    KEY_SEQUENCE = 'sequence'

    def __init__(self, vocab_size: int, embedding: nn.Embedding,
                    max_len: int, embedding_size: int, sos_id,
                    eos_id, config, n_layers: int = 1, 
                    rnn_cell: str = 'gru',
                    bidirectional: bool = False, 
                    input_dropout_perc: float = 0.,
                    dropout_perc: float = 0.):
        hidden_size = embedding_size
        if bidirectional:
            hidden_size *= 2

        super().__init__(vocab_size, max_len, hidden_size,
                            input_dropout_perc, dropout_perc,
                            n_layers, rnn_cell)
        
        self.bidirectional_encoder = bidirectional
        self.rnn = self.rnn_cell(embedding_size, hidden_size, 
                                    n_layers, batch_first = True, 
                                    dropout = dropout_perc)
        self.output_size = vocab_size
        self.max_length = max_len
        self.eos_id = eos_id
        self.sos_id = sos_id
        self.config = config
        self.init_input = None
        self.embedding = embedding
        
        self.attention_title = Attention(self.hidden_size)
        self.attention_hidden = Attention(self.hidden_size)

        self.fc = Gate(self.hidden_size)
        self.out1 = nn.Linear(self.hidden_size, self.output_size)
        self.out2 = nn.Linear(self.hidden_size, self.output_size)

    def _step(self, input_var, pg_encoder_states, hidden,
                    encoder_outputs, function):
        batch_size = input_var.size(0)
        output_size = input_var.size(1)

        embedded = self.embedding(input_var)
        embedded = self.input_dropout(embedded)

        output_states, hidden = self.rnn(embedded, hidden)

        attn = None

        output_states_attn1, attn1 = self.attention_title(output_states, 
                                                          encoder_outputs)

        if pg_encoder_states is None:
            output_states_attn = output_states_attn1
        else:
            output_states_attn2, attn2 = self.attention_hidden(output_states, 
                                                               pg_encoder_states)
            output_states_attn = self.fc(output_states_attn1, 
                                         output_states_attn2)

        outputs = self.out1(output_states_attn.view(-1, self.hidden_size)).view(batch_size, output_size, -1)
        return outputs, output_states_attn, hidden, attn
    
    def _validate_args(self, inputs, encoder_hidden, encoder_outputs, 
                        function, teacher_forcing_ratio):
        if encoder_outputs is None:
            raise ValueError("Argument encoder_outputs cannot be None when attention is used.")

        # inference batch size
        if inputs is None and encoder_hidden is None:
            batch_size = 1
        else:
            if inputs is not None:
                batch_size = inputs.size(0)
            else:
                if self.rnn_cell is nn.LSTM:
                    batch_size = encoder_hidden[0].size(1)
                elif self.rnn_cell is nn.GRU:
                    batch_size = encoder_hidden.size(1)

        if inputs is None:
            if teacher_forcing_ratio > 0:
                raise ValueError("Teacher forcing has to be disabled (set 0) when no inputs is provided.")
            inputs = torch.LongTensor([self.sos_id] * batch_size).view(batch_size, 1)
            if self.config.cuda_is_available:
                inputs = inputs.cuda()
            max_length = self.max_length
        else:
            max_length = inputs.size(1) - 1   

        return inputs, batch_size, max_length

    def _init_state(self, encoder_hidden):
        if encoder_hidden is None:
            return None
        if isinstance(encoder_hidden, tuple):
            encoder_hidden = tuple([self._cat_directions(h) for h in encoder_hidden])
        else:
            encoder_hidden = self._cat_directions(encoder_hidden)
        return encoder_hidden

    def _cat_directions(self, h):
        if self.bidirectional_encoder:
            h = torch.cat([h[0:h.size(0):2], h[1:h.size(0):2]], 2)
        return h

    def forward(self, inputs = None, encoder_hidden = None, 
                    encoder_outputs = None, pg_encoder_states = None,
                    function = F.log_softmax, teacher_forcing_ratio = 0):
        ret_dict = dict()
        ret_dict[DecoderRNNFB.KEY_ATTN_SCORE] = list()

        inputs, batch_size, max_length = self._validate_args(inputs,
                                    encoder_hidden, encoder_outputs,
                                    function, teacher_forcing_ratio)
        
        decoder_hidden = self._init_state(encoder_hidden)
        use_teacher_forcing = True if teacher_forcing_ratio == 1 else False

        if use_teacher_forcing:
            decoder_input = inputs[:, :-1]
            decoder_outputs, decoder_output_states, decoder_hidden, attn = \
                self._step(decoder_input, pg_encoder_states,
                                decoder_hidden, encoder_outputs, function=function)   
        else:
            decoder_outputs = []
            decoder_output_states = []
            sequence_symbols = []
            lengths = np.array([max_length] * batch_size)    

            def decode(step, step_output, step_output_state=None, step_attn=None):
                if step_output_state is not None:
                    decoder_outputs.append(step_output)
                    decoder_output_states.append(step_output_state)
                ret_dict[DecoderRNNFB.KEY_ATTN_SCORE].append(step_attn)
                symbols = step_output.topk(1)[1]
                sequence_symbols.append(symbols)

                eos_batches = symbols.data.eq(self.eos_id)
                if eos_batches.dim() > 0:
                    eos_batches = eos_batches.cpu().view(-1).numpy()
                    update_idx = ((lengths > step) & eos_batches) != 0
                    lengths[update_idx] = len(sequence_symbols)
                return symbols         

            decoder_input = inputs[:, 0].unsqueeze(1)
            for di in range(max_length):                            
                decoder_output, decoder_output_state, decoder_hidden, step_attn = \
                    self.forward_step(decoder_input, pg_encoder_states, decoder_hidden,
                                      encoder_outputs, function=function)
                # # not allow decoder to output UNK
                decoder_output[:, :, 3] = -float('inf')

                step_output = decoder_output.squeeze(1)
                step_output_state = decoder_output_state.squeeze(1)
                symbols = decode(di, step_output, step_output_state, step_attn)
                decoder_input = symbols

            decoder_outputs = torch.stack(decoder_outputs, dim=1)
            decoder_output_states = torch.stack(decoder_output_states, dim=1)
            ret_dict[DecoderRNNFB.KEY_SEQUENCE] = sequence_symbols
            ret_dict[DecoderRNNFB.KEY_LENGTH] = lengths.tolist()

        return decoder_outputs, decoder_output_states, ret_dict                


import logging
from ..static import LOGGER_PRINT_LEVEL
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(LOGGER_PRINT_LEVEL)

import torch.nn as nn

class BaseRNN(nn.Module):
    def __init__(self, vocab_size: int, max_len: int, hidden_size: int, 
                    input_dropout_perc: float, dropout_perc: float,
                    n_layers: int, rnn_cell: str) -> None:
        super().__init__()
        self.vocab_size = vocab_size
        self.max_len = max_len
        self.hidden_size = hidden_size
        self.n_layers = n_layers
        self.input_dropout_perc = input_dropout_perc
        self.dropout_perc = dropout_perc
        
        rnn_cell = rnn_cell.lower()

        if rnn_cell == 'lstm':
            self.rnn_cell = nn.LSTM
        elif rnn_cell == 'gru':
            self.rnn_cell = nn.GRU
        else:
            raise ValueError('unsupported RNN cell: {rnn_cell}'.format(rnn_cell=rnn_cell))
        
        self.input_dropout = nn.Dropout(p=input_dropout_perc)

    def forward(self, *args, **kwargs) -> None:
        raise NotImplementedError()
    


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='project_6127',
      version='0.0.04',
      description='Package for AI6127 Project',
      long_description=long_description,
      url='https://github.com/AiRiFiEd/project_6127',
      author='yqlim',
      author_email='yuanqing87@gmail.com',
      license='MIT',
      packages=find_packages(),
      zip_safe=False)


import logging
from ..static import LOGGER_PRINT_LEVEL
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(LOGGER_PRINT_LEVEL)

import torch.nn as nn

from .base import BaseRNN

class EncoderRNN(BaseRNN):
    def __init__(self, vocab_size: int, embedding: nn.Embedding,
                    max_len: int, hidden_size: int, 
                    input_dropout_perc: float = 0.,
                    dropout_perc: float = 0., n_layers: int = 1, 
                    bidirectional: bool = False, rnn_cell: str = 'gru',
                    variable_lengths: bool = True) -> None:
        super().__init__(vocab_size=vocab_size, 
                            max_len=max_len,
                            hidden_size=hidden_size,
                            input_dropout_perc=input_dropout_perc,
                            dropout_perc=dropout_perc,
                            n_layers=n_layers, rnn_cell=rnn_cell)
        self.embedding = embedding
        self.variable_lengths = variable_lengths
        self.rnn = self.rnn_cell(
            hidden_size, hidden_size, n_layers,
            batch_first=True, bidirectional=bidirectional, 
            dropout=dropout_perc
        )

    def forward(self, input_var, input_lengths=None):
        embedded = self.embedding(input_var)
        embedded = self.input_dropout(embedded)
        if self.variable_lengths:
            try:
                embedded = nn.utils.rnn.pack_padded_sequence(embedded, 
                                input_lengths, batch_first=True)
            except:
                print(input_lengths)
        output, hidden = self.rnn(embedded)
        if self.variable_lengths:
            output, _ = nn.utils.rnn.pad_packed_sequence(output, batch_first=True)

        return output, hidden